# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 07/01/2020.

from pytonik.Functions.validation import validation

class readmore(validation):


    def __getattr__(self, item):
        return item

    def __call__(self, *args, **kwargs):

        return None

    def __init__(self, *args,  **kwargs):

        if len(args) > 0 or len(kwargs) > 0:
            if all(args) is not False:
                self.strt = self.lstring(*args, **kwargs)
            else:
                self.strt = self.lstring(**kwargs)

        return None

    def __str__(self):

        return self.strt


    def lstring(self, text="", trim = 'False', length = '10000000000000', link="", label="Read more",  css="readmore"):

        if bool(trim) is True:
            new_text = self.trim(text)

        else:
            new_text = text
        readlable  = " <a class='{css}' href='{link}'> {label} </a>".format(css=css, link=link, label=label)
        data = self.limit(self, text=new_text, trim = trim, length = length, label=readlable)

        return data

    def limit(self, text="", trim = 'False', length = '10000000000000', label=""):

        if bool(trim) is True:
            new_text = self.trim(text)

        else:
            new_text = text

        data = (new_text[:int(length)] + str(label)) if len(new_text) > int(length) else new_text

        return data


import cgitb
import base64
import traceback
import cgi
import ssl
import argparse
import locale
import os
import re
import sys
import time
import json
import warnings
from pytonik.cmd import lang
from pytonik.Version import *
from pytonik.util.Variable import Variable
from typing import Any, Callable, Dict, List, Pattern, Union
from pytonik.cmd.console import (  # type: ignore
    colorize, bold, red, green, turquoise, nocolor, color_terminal
)

try:
    from http import cookies as cook
except Exception as err:
    import Cookie as cook

from socketserver import ThreadingMixIn

try:
    from BaseHTTPServer import BaseHTTPRequestHandler
except Exception as err:
    from http.server import BaseHTTPRequestHandler, HTTPServer
try:
    import socket
except Exception as err:
    from socket import socket

try:
    import http.client as htp
    from http import client
    import urllib as urllib
except Exception as err:
    import httplib as htp
    import urllib2 as urllib

varb = Variable()


def run(host="", path="", port=6060, server_pro="HTTP/1.1", ssl_ip="", ssl_port="", pr=False):
    server = HTTPServer

    path = str(path).replace(
        "\\", "/") if path != "" else str(os.getcwd()).replace("\\", "/")

    spes = "/"
    vpath = ""

    try:
        import imp as im
    except Exception as err:
        import importlib as im

    mimetype = ""
    sys.path.insert(0, os.path.dirname(__file__))
    os.chdir(path)

    cookie_v = cook

    class httpv(BaseHTTPRequestHandler):

        def do_GET(self):

            path_info = self.path
            
            
            form = cgi.FieldStorage(
                fp=self.rfile,
                headers=self.headers,
                environ={
                    'REQUEST_METHOD': "GET",
                    'CONTENT-TYPE': self.headers['Centent-Type']
                }
            )
            if self.path == spes:
                if os.path.isfile(str(path) + spes + "public" + spes + "index.py") == True:
                    vpath = "public" + spes + "index.py"

                elif os.path.isfile(str(path) + spes + "public" + spes + "home.py") == True:
                    vpath = "public" + spes + "home.py"
                elif os.path.isfile(str(path) + spes + "public" + spes + "default.py") == True:
                    vpath = "public" + spes + "default.py"

                App = im.load_source('App.App', path + spes + vpath)
                mimetype = 'text/html'
                if "." not in path_info :
                    
                    
                    App.App.put(method="GET", accept_lang=self.headers["Accept-Language"],
                            http_connect=self.headers["Connection"], http_user_agt=self.headers["User-Agent"],
                            http_encode=self.headers["Accept-Encoding"], path=path, host=host, port=port,
                            para=path_info, remoter_addr=self.client_address[
                                0], remoter_port=self.client_address[1],
                            script_file=str(
                                path) + str(spes) + (vpath), server_proto=server_pro, server_ver=self.server_version,
                            protocol_ver=self.protocol_version)
                    runs_response = App.App.runs(formData=form)
                    if isinstance(runs_response, tuple) == True:

                        if str(runs_response[0]) == "404" or str(runs_response[0]) == "405" or str(
                                runs_response[0]) == "400":

                            self.error(runs_response[0], runs_response[1])

                        elif str(runs_response[0]) == "307":

                            self.redirect(runs_response[0], runs_response[1])
                        else:

                            self.rendering(mimetype=mimetype,
                                        content=runs_response)

                    else:
                        self.rendering(mimetype=mimetype, content=runs_response)

            elif self.path != spes:

                if "." not in str(self.path):

                    if str(self.path) != "":
                        if os.path.isfile(str(path) + spes + "public" + spes + "index.py") == True:
                            vpath = "public" + spes + "index.py"

                        elif os.path.isfile(str(path) + spes + "public" + spes + "home.py") == True:
                            vpath = "public" + spes + "home.py"

                        elif os.path.isfile(str(path) + spes + "public" + spes + "default.py") == True:

                            vpath = "public" + spes + "default.py"

                        App = im.load_source('App.App', path + spes + vpath)
                        mimetype = 'text/html'
                        if "." not in path_info:
                            
                            
                            App.App.put(method="GET", accept_lang=self.headers["Accept-Language"],
                                        http_connect=self.headers["Connection"], http_user_agt=self.headers["User-Agent"],
                                        http_encode=self.headers["Accept-Encoding"], path=path, host=host, port=port,
                                        para=path_info, remoter_addr=self.client_address[0],
                                        remoter_port=self.client_address[1], script_file=str(
                                path) + str(spes) + (vpath), server_proto=server_pro, server_ver=self.server_version,
                                protocol_ver=self.protocol_version)

                            runs_response = App.App.runs(formData=form)
                            if isinstance(runs_response, tuple) == True:

                                if str(runs_response[0]) == "404" or str(runs_response[0]) == "405" or str(
                                        runs_response[0]) == "400":

                                    self.error(runs_response[0], runs_response[1])

                                elif str(runs_response[0]) == "307":

                                    self.redirect(
                                        runs_response[0], runs_response[1])
                                else:

                                    self.rendering(mimetype=mimetype,
                                                content=runs_response)

                            else:
                                self.rendering(mimetype=mimetype,
                                            content=runs_response)

            if self.path.endswith('favicon.ico'):
                return
            try:
                for mime in MIME_TYPES:
                    if self.path.endswith(mime['ext']):
                        self.rendering(
                            path=path, mimetype=mime['type'], mode=mime['mode'], code=200)
                    else:
                        split_path = str(self.path).split('?')
                        if split_path[0].endswith(mime['ext']):
                            self.rendering(
                            path=path, mimetype=mime['type'], mode=mime['mode'], code=200)


            except Exception as err:
                if self.path.endswith(self.path):
                    if os.path.isfile(str(path) + spes + "public" + spes + "index.py") == True:
                        vpath = "public" + spes + "index.py"

                    elif os.path.isfile(str(path) + spes + "public" + spes + "home.py") == True:
                        vpath = "public" + spes + "home.py"

                    elif os.path.isfile(str(path) + spes + "public" + spes + "default.py") == True:
                        vpath = "public" + spes + "default.py"

                    App = im.load_source('App.App', path + spes + vpath)
                    code = "500"
                    App.App.put(status=code)
                    pth = str(os.path.dirname(
                        os.path.abspath(__file__))).replace("\\", "/")
                    f = open(pth + "/cmd/errd/index.html", "r")
                    content = str(f.read()).format(code=code, name=AUTHOR, message=HTTP_CODE.get(
                        code, ""), version=VERSION_TEXT)
                    self.wfile.write(bytes(str(content).encode()))

        def do_POST(self):
            path_info = self.path
           
            form = cgi.FieldStorage(
                fp=self.rfile,
                headers=self.headers,
                environ={
                    'REQUEST_METHOD': "POST",
                    'CONTENT-TYPE': self.headers['Content-Type']
                }
            )

            # text/plain; charset=utf-8
            if os.path.isfile(str(path) + spes + "public" + spes + "index.py") == True:
                vpath = "public" + spes + "index.py"

            elif os.path.isfile(str(path) + spes + "public" + spes + "home.py") == True:
                vpath = "public" + spes + "home.py"

            else:
                vpath = "public" + spes + "default.py"

            App = im.load_source('App.App', path + spes + vpath)
            mimetype = 'text/html'
            if "." not in path_info :
                
                
                App.App.put(method="POST", accept_lang=self.headers["Accept-Language"],
                            http_connect=self.headers["Connection"], http_user_agt=self.headers["User-Agent"],
                            http_encode=self.headers["Accept-Encoding"], path=path, host=host, port=port, para=path_info,
                            remoter_addr=self.client_address[0], remoter_port=self.client_address[1], script_file=str(
                    path) + str(spes) + (vpath), server_proto=server_pro, server_ver=self.server_version,
                    protocol_ver=self.protocol_version)

                self.rendering(mimetype=mimetype, code=200,
                            content=App.App.runs(formData=form))

        def do_HEAD(self):
            self.do_GET()

        def do_PUT(self):
            self.do_POST()

        def rendering(self, path="", mimetype="", mode='r', encoding="utf-8", content="", code=200):

            self.send_response(int(code))
            self.send_header('Content-type', mimetype)
            self.end_headers()
            if path != "":

                f = open(path + self.path, mode)
                readv = ""
                if mode == "rb":
                    readv = f.read()
                else:
                    readv = bytes(str(f.read()).encode('utf-8'))

                self.wfile.write(readv)

                f.close()

            elif content != "":
                self.wfile.write(bytes(str(content).encode()))

        def error(self, code, e_url, code_re=307):
            self.send_response(int(code_re))
            self.send_header('Location', "{e_url}".format(e_url=e_url))
            self.send_error(
                code=int(code), message=HTTP_CODE.get(code, ""))
            self.end_headers()

        def redirect(self, code, re_url, code_re=307):
            self.send_response(int(code_re))
            self.send_header('Location', "{re_url}".format(re_url=re_url))
            self.send_error(
                code=int(code), message=HTTP_CODE.get(code, ""))
            self.end_headers()

    class ThreadedHTTPServer(ThreadingMixIn, server):
        """Moomins live here"""

    hostname = ssl_ip if ssl_ip != "" else host
    portnumber = int(ssl_port) if ssl_port != "" else int(port)
    vars_http = ""
    try:
        context = ssl.create_default_context()

        with socket.create_connection((hostname, port)) as sock:
            with context.wrap_socket(sock, server_hostname=hostname) as ssock:
                # print(ssock.version())
                data = json.dumps(ssock.getpeercert())
        varb.put("HTTPS", "on")
        vars_http = "https://"
        # print(ssock.getpeercert())
    except Exception as err:
        try:
            cert = ssl.get_server_certificate((hostname, int(portnumber)))
            varb.put("HTTPS", "on")
            vars_http = "https://"
        except Exception as err:
            varb.put("HTTPS", "off")
            vars_http = "http://"

    try:
        l = host if str(port) == "8080" or str(port) == "80" else "{}:{}".format(
            host, port)
        if pr == True:

            print(
                bold(green("Pytonik development server running on " + str(vars_http) + str(l))))
        else:
            print(bold(green("Pytonik server running on " + str(vars_http) + str(l))))

        server = ThreadedHTTPServer((host, port), httpv)
        server.serve_forever()
        server.server_close()
    except Exception as err:
        print(bold(red("Something went wrong: Default port already in use")))



# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 13/01/2020.
__path__ = __import__('pkgutil').extend_path(__path__, __name__)

from pytonik.cmd import start
from pytonik.cmd import server
from pytonik.cmd import doc
from pytonik.cmd import install 
from pytonik.cmd import help


# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 3/4/20.
import os, sys
from pytonik import Version


class Variable:

    def __getattr__(self, item):
        return item

    def __init__(self):
        
        return None

    def put(self, accept_lang="", http_user_agt="", http_encode="",  http_connect="", path="", host="", port="", para="", status=200, accept="", method="", referral="", remoter_addr="", remoter_port="", script_file="", server_proto="", server_ver="", protocol_ver="", redirect_url="", ssl_v=""):
        
        uri, query_string = "", ""  # path.split('/')[-1]+str(para)
        
        host = str(host) if str(host) != "" else os.environ.get("HTTP_HOST", "")
        para_v = "/"+str(os.path.basename(os.getcwd()))+str(para) if str(para) != "" else os.environ.get("REQUEST_URI", "")
        
        port = str(port) if str(port) != "" else os.environ.get("SERVER_PORT", "")
        
        path = str(path) if str(path) != "" else os.environ.get("PATH_INFO", "")
        
        accept_lang = str(accept_lang) if str(accept_lang) != "" else os.environ.get("HTTP_ACCEPT_LANGUAGE", "")
        
        http_connect  = str(http_connect) if str(http_connect) != "" else os.environ.get("HTTP_CONNECTION", "")
        
        http_user_agt = str(http_user_agt) if str(http_user_agt) != "" else os.environ.get("HTTP_USER_AGENT", "")
        
        http_encode = str(http_encode) if str(http_encode) != "" else os.environ.get("HTTP_ACCEPT_ENCODING", "")
    
        ssl_v = str(ssl_v) if str(ssl_v) != "" else os.environ.get("HTTPS", "")
        
        
        status = str(status) if status != "" else os.environ.get("REDIRECT_STATUS", "")
        status = str(status) if status != "" else os.environ.get("REDIRECT_REDIRECT_STATUS", "")

        http_status = str(status) if status != "" else os.environ.get("HTTP_STATUS", "")
        
        remoter_addr = str(remoter_addr) if remoter_addr != "" else os.environ.get("REMOTE_ADDR", "")
        
        remoter_port = str(remoter_port) if remoter_port != "" else os.environ.get("REMOTE_PORT", "")
         
        script_file = str(script_file) if script_file != "" else os.environ.get("SCRIPT_FILENAME", "")
    
        server_ver = str(server_ver) if server_ver != "" else os.environ.get("SERVER_VERSION", "")

        v_referral = ""
        if os.environ.get("HTTP_REFERER", "") != "":
            if para == referral:
                v_referral = ""
            else:
                v_referral = referral if referral != "" else os.environ.get("HTTP_REFERER", "")
        else:
            v_referral = para
       
        server_proto = str(server_proto) if str(
            server_proto) != "" else os.environ.get("SERVER_PROTOCOL", "")
      
        protocol_ver = str(protocol_ver) if str(
            protocol_ver) != "" else os.environ.get("PROTOCOL_VERSION", "")
        
        
        
        redirect_url = str(redirect_url) if str(redirect_url) != "" else os.environ.get("REDIRECT_URL", "")
      
        if para.find("?") == True:
            parts = para.split('?')
            if len(parts) > 1:
                uri = parts[0]
                query_string = parts[1]
            else:
                uri = ""
                query_string = parts[0]

        else:
            query_string = ""
            uri = ""
        server_software = Version.AUTHOR+"/"+Version.VERSION_TEXT if Version.AUTHOR != "" else os.environ.get("SERVER_SOFTWARE", "")
        env = {
            "Framework": Version.AUTHOR if Version.AUTHOR != "" else os.environ.get("Framework", ""),
            "X-Version": Version.VERSION_TEXT if Version.VERSION_TEXT != "" else os.environ.get("X-Version", ""),
            "X-Organisation": Version.ORG if Version.ORG != "" else os.environ.get("X-Organisation", ""),
            "PATH_INFO": para,
            "HTTP_HOST": host.replace(":80", ""),
            "HTTP_STATUS": http_status,
            "HTTPS": ssl_v,
            "SERVER_PORT": port,
            "REDIRECT_STATUS": status,
            "REDIRECT_REDIRECT_STATUS": status,
            "REMOTE_ADDR": remoter_addr,
            "REMOTE_PORT": remoter_port,
            "SCRIPT_FILENAME": script_file,
            "HTTP_CONNECTION": http_connect, 
            "HTTP_ACCEPT_LANGUAGE": accept_lang,
            "HTTP_ACCEPT_ENCODING": http_encode,
            "HTTP_USER_AGENT": http_user_agt,
            "HTTP_ACCEPT": str(accept) if accept != "" else os.environ.get("HTTP_ACCEPT", ""),
            "DOCUMENT_ROOT": path,
            "REQUEST_METHOD": str(method) if method != "" else os.environ.get("REQUEST_METHOD", ""),
            "QUERY_STRING": str(query_string),
            'SERVER_SOFTWARE': Version.AUTHOR,
            'PATH_TRANSLATED': str(path) + "/public",
            'SERVER_SIGNATURE':  "{server_software} {sys_os} PYTHON/{py_version} Server at {host} Port {port}".format(py_version=str(sys.version_info.major) + "."+ str(sys.version_info.minor) + "."+ str(sys.version_info.micro), sys_os=sys.platform, server_software=server_software,host=host, port=port),
            'HTTP_REFERER': v_referral,
            'SERVER_PROTOCOL': server_proto,
            'SERVER_VERSION': server_ver,
            'REDIRECT_URL': redirect_url,
            'PROTOCOL_VERSION': protocol_ver,
            "REQUEST_URI": para_v
        }
        
        
        if Version.PYVERSION_MA <= 2:
            lt = env.iteritems()
        else:
            lt = env.items()
        if os.environ.get("REQUEST_URI", "") != "" or os.environ.get("REQUEST_URI", "") != None:
            #os.environ.clear()
            os.environ.update(dict(env))
        
        for k, v in lt:
            if k not in os.environ:
                self.default(str(k), str(v).encode())
            else:
                upd = {str(k): str(v)}
                self.update(dict(upd))
            

    def out(self, key, alt=""):
        return os.environ.get(key, alt)

    def see(self):
        return os.environ
    
        
    def default(self, key, value):
        if key in os.environ:
            upd = {str(key) : str(value)}
            return self.update(dict(upd))
        else:
            return os.environ.setdefault(str(key), str(value))
    
              
    def update(self, value=dict):
        return os.environ.update(value)
    
        


# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 07/11/2019.

from pytonik.Lang import Lang
from pytonik.Controllers import Controllers


class __():

    def __getattr__(self, item):
        return item

    def __call__(self, *args, **kwargs):

        return None

    def __init__(self, *args, **kwargs):
        self.langs = Lang(Controllers().languages)
        self.langs.loadLang()
        if len(args) > 0 or len(kwargs) > 0:
            if all(args) != False:
                self.lg = self.lang(*args, **kwargs)
            else:
                self.lg = self.lang(**kwargs)
		
        return None

    def __str__(self):

        return self.lg


    def lang(self, lang="", defindValue = ''):
        if lang != "":
            return self.langs.get(lang, defindValue)



    def default(self):

        return self.langs.lg



###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###


import smtplib, string, random
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email import encoders
import os

from pytonik import Router, Config, Log
log_msg = Log.Log()


class SMTP():

    def __init__(self):

        self.Router = Router.Router()
        self.envSMTP = self.Router.env()

        Conf = Config.Config()

        Conf.add(self.envSMTP)
        setting = Conf.get('SMTP')
        self.server = setting.get('server', '')
        self.port = setting.get('port', '')
        self.username = setting.get('username', '')
        self.password = setting.get('password', '')

        self.con = None
        self.result = None
        self.response = None
        self.attachfile = None
        self.connect()
        return None

    def connect(self):
        try:
            self.response = smtplib.SMTP(self.server, self.port)
            self.response.starttls()
            self.response.login(self.username, self.password)
            return self.response
        except Exception as err:
            log_msg.error(err)


    def close(self):
        return self.response.quit()


    def send(self, from_send, to_recipient, message_subject= "", messege_content = "", header="html"):
        try:
            body = messege_content
            msg = MIMEMultipart()
            msg['From'] = from_send
            msg['To'] = to_recipient
            msg['Subject'] = message_subject
            msg.attach(MIMEText(body, header))
            if self.attachfile != None:
                msg.attach(self.attachfile)

            context = msg.as_string()
            self.response.sendmail(from_send, to_recipient, context)
            self.close()
            return True
        except Exception as err:
            log_msg.error(err)
            return err

    def attach(self, atfile = "", rename=""):
        if atfile != "":
            try:
                ext = os.path.splitext(atfile)[1][1:]
                filename = atfile if rename == "" else str(rename) + '.' + str(ext)
                attach_file = open(atfile, 'rb')
                attach_load = MIMEBase('application', 'octet-stream')
                attach_load.set_payload(attach_file.read())
                encoders.encode_base64(attach_load)
                attach_load.add_header('Content-Disposition', 'attachment', filename=filename)
                self.attachfile = attach_load

            except Exception as err:
                try:
                    ext =  os.path.splitext(atfile.filename)[1][1:]
                    filename = atfile.filename if rename == "" else str(rename)+'.'+str(ext)
                    attach_load = MIMEBase('application', 'octet-stream')
                    read_attach_file = (atfile.file).read()
                    attach_load.set_payload(read_attach_file)
                    encoders.encode_base64(attach_load)
                    attach_load.add_header('Content-Disposition', 'attachment',  filename=filename)

                    self.attachfile = attach_load

                except Exception as err:

                    log_msg.error(err)

        return self


    def retrieve(self):
        return self.response.getreply()


###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###


import sys
import os
import re
from pytonik import Log
from pytonik.Version import *
from pytonik.Config import Config
from pytonik.Core.env import env
from pytonik.Core import Helpers
from pytonik.Controllers import Controllers
from pytonik.util.Variable import Variable

log_msg = Log.Log()


class Router(Controllers):
    def args(self, to="", params=[], link=False, code=200):
        return to, params, link, code

    def __getattr__(self, item):
        return item

    def __init__(self):
        
        self.control = Controllers()
        self._route = []
        self._params = {}
        self._getcontrol = []
        self._getaction = []
        self._geturi = []
        self._redirect = []
        self._method = []
        self.route_to = []
        self._getmethod = []
        self._despatch = []
        self._code = []
        self._error = []
        self._geterrocontrol = []
        self._link = []
        self._getparams = []
        self.objv = ""
        
        return None

    def get(self, uri, call=""):
        
        if call != "":
            if isinstance(call, tuple):
                self._route_(route=uri, call=call, method="GET")
            else:
                
                lcall = []
                method_v, control_v = "", ""

                if len(call.split('@')) > 0:
                    control = call.split('@')
                    if len(control) == 2:
                        control_v = str(control[0]).lower().replace(
                            'Controller', '').replace('controller', '')
                        if len(str(control[1]).split(":")) > 0:
                            method_v = control[1].split(":")[0]
                            param = control[1].split(":")[1:]
                            lcall.append(uri)
                            lcall.append(param)
                        else:
                            method_v = control[1]
                    else:

                        if control[0] == "/":
                            control_v = self.control.default_controllers
                        else:
                            control_v = str(control[0]).lower().replace(
                                'Controller', '').replace('controller', '')

                        method_v = self.control.default_actions
                    lroute = str(control_v)+"/"+str(method_v)

                    self._route_(route=lroute, call=lcall, method="GET")

        return self



    def post(self, uri, call=""):

        
        if call != "":
            if isinstance(call, tuple):
                self._route_(route=uri, call=call, method="POST")
            else:
                lcall = []
                method_v, control_v = "", ""
                if len(call.split('@')) > 0:
                    control = call.split('@')
                    if len(control) == 2:
                        control_v = str(control[0]).lower().replace(
                            'Controller', '').replace('controller', '')
                        if len(str(control[1]).split(":")) > 0:
                            method_v = control[1].split(":")[0]
                            param = control[1].split(":")[1:]
                            lcall.append(uri)
                            lcall.append(param)
                        else:
                            method_v = control[1]
                    else:
                        if control[0] == "/":
                            control_v = self.control.default_controllers
                        else:
                            control_v = str(control[0]).lower().replace(
                                'Controller', '').replace('controller', '')

                        method_v = self.control.default_actions
                    lroute = str(control_v)+"/"+str(method_v)

                    self._route_(route=lroute, call=lcall, method="POST")
        return self

    def put(self, uri, call=""):
        if call != "":
            if isinstance(call, tuple):
                self._route_(route=uri, call=call, method="POST")
            else:
                lcall = []
                method_v, control_v = "", ""
                if len(call.split('@')) > 0:
                    control = call.split('@')
                    if len(control) == 2:
                        control_v = str(control[0]).lower().replace(
                            'Controller', '').replace('controller', '')
                        if len(str(control[1]).split(":")) > 0:
                            method_v = control[1].split(":")[0]
                            param = control[1].split(":")[1:]
                            lcall.append(uri)
                            lcall.append(param)
                        else:
                            method_v = control[1]
                    else:
                        if control[0] == "/":
                            control_v = self.control.default_controllers
                        else:
                            control_v = str(control[0]).lower().replace(
                                'Controller', '').replace('controller', '')

                        method_v = self.control.default_actions
                    lroute = str(control_v)+"/"+str(method_v)

                    self._route_(route=lroute, call=lcall, method="POST")
        return self

    def any(self, uri, call=""):
        if call != "":
            if isinstance(call, tuple):
                self._route_(route=uri, call=call)
            else:
                lcall = []
                method_v, control_v = "", ""
                if len(call.split('@')) > 0:
                    control = call.split('@')
                    if len(control) == 2:
                        control_v = str(control[0]).lower().replace(
                            'Controller', '').replace('controller', '')
                        if len(str(control[1]).split(":")) > 0:
                            method_v = control[1].split(":")[0]
                            param = control[1].split(":")[1:]
                            lcall.append(uri)
                            lcall.append(param)
                        else:
                            method_v = control[1]
                    else:
                        if control[0] == "/":
                            control_v = self.control.default_controllers
                        else:
                            control_v = str(control[0]).lower().replace(
                                'Controller', '').replace('controller', '')

                        method_v = self.control.default_actions
                    lroute = str(control_v)+"/"+str(method_v)
                    
                    self._route_(route=lroute, call=lcall)
        return self

    def error(self, uri, call="", code=302):
        if isinstance(call, tuple):
            if call[0] == "/":
                replace = self.control.default_controllers
            else:
                replace = call[0]
            code = code if int(call[3]) == int(200) else call[3]
        else:
            if call[0] == "/":
                replace = self.control.default_controllers
            else:
                replace = call
            code = int(code)

        self._geterrocontrol.append(uri)
        self._link.append(call[2])
        self._error.append(replace)
        self._code.append(code)

        return self

    def redirect(self, uri, call="", code=302):

        if isinstance(call, tuple):
            if call[0] == "/":
                replace = self.control.default_controllers
            else:
                replace = call[0]
            code = code if int(call[3]) == int(200) else call[3]
        else:
            if call == "/":
                replace = self.control.default_controllers
            else:
                replace = call
            code = int(code)

        self._despatch.append(uri)
        self._redirect.append(replace)
        self._code.append(code)

        return self

    def permanentRedirect(self, uri, call="", code=301):
        if isinstance(call, tuple):
            if call[0] == "/":
                replace = self.control.default_controllers
            else:
                replace = call[0]
            code = code if int(call[3]) == int(200) else call[3]
        else:
            if call == "/":
                replace = self.control.default_controllers
            else:
                replace = call
            code = int(code)

        self._despatch.append(uri)
        self._redirect.append(replace)
        self._code.append(code)
        return self

    def where(self, *args):

        if len(args) > 0:

            if isinstance(args[0], dict):
                if PYVERSION_MA >= 3:

                    params = args[0].items()
                else:
                    params = args[0].iteritems()

                for k, v in params:
                    regex = re.compile(v)
                    if regex.match(self._params.get(k, "")):
                        return True
                    else:
                        return False

            elif isinstance(args, list):

                regex = re.compile(args[1])
                if regex.match(self._params.get(args[0], "")):
                    return True
                else:
                    return False

        else:
            return False

    def _route_(self, route="", call="", method=""):
        
        route = route.split('/')
        new_para = route

        parameter = {}
        replace = ""
        new_uri = []
        list_params = []
        params = ""

        new_paraf = []
        if len(call) > 0:
            replace = call[0]
            #if len(call[1]) > 0:
            self._getparams.append(call[1])
            
            #if len(replace) > 0:
            self._route.append(replace)

        if len(route) == 1:
            ctl = route[0] if route[0] != "" else self.control.default_controllers
            self._getcontrol.append(ctl)

            self._getaction.append(self.control.default_actions)
        elif len(route) > 1:
            self._getcontrol.append(
                route[0] if route[0] != "" else self.control.default_controllers)
            self._getaction.append(
                route[1] if route[1] != "" else self.control.default_actions)

        self._method.append(method)    


    def getParams(self):
        return self._getparams

    def getController(self):
        return self._getcontrol

    def getAction(self):
        return self._getaction

    def getRouter(self):
        return self._route

    def getMethod(self):
        return self._method

    def getRediret(self):
        return self._redirect

    def getDespatch(self):
        return self._despatch

    def getCode(self):
        return self._code

    def getError(self):
        return self._error

    def getLink(self):
        return self._link

    def getErrorControl(self):
        return self._geterrocontrol



# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 2/24/20.

from pytonik.Log import Log

class LogError(Exception):
    
    def __getattr__(self, item):
        return item

    def __init__(self, error=""):
         self.error = error
         
    def __str__(self):
        return self.error


class TemplateError(Exception):
    pass


class TemplateContextError(TemplateError):

    def __init__(self, context_var):
        self.context_var = context_var


    def __str__(self):
        Log('').error("cannot resolve '{}'".format(self.context_var))
        return "cannot resolve '{}'".format(self.context_var)


class TemplateSyntaxError(TemplateError):

    def __init__(self, error_syntax):
        self.error_syntax = error_syntax

    def __str__(self):
        Log('').error(" '{}' seems like invalid syntax".format(self.error_syntax))
        return "'{}' seems like invalid syntax".format(self.error_syntax)




from pytonik.Editor._Fragment import _Fragment
from pytonik.util.Exception import TemplateSyntaxError, TemplateError, TemplateSyntaxError, TemplateContextError
from pytonik.Editor.Node import _Node, _Call,  _Root, _ScopableNode, _Text, _Variable, _If, _Each, _Else, _Block
from pytonik.Editor.Resolver import *

class Compiler(object):
    def __init__(self, template_string):
        self.template_string = template_string

    def each_fragment(self):
        
        for fragment in TOK_REGEX.split(self.template_string):
            
            if fragment:
                
                yield _Fragment(fragment)

    def compile(self):

        root = _Root()
        scope_stack = [root]
        
        
        for fragment in self.each_fragment():

            if not scope_stack:
                raise TemplateError('nesting issues')

            parent_scope = scope_stack[-1]


            if fragment.type == CLOSE_BLOCK_FRAGMENT:
                parent_scope.exit_scope()
                scope_stack.pop()
                continue

            new_node = self.create_node(fragment)

            if new_node:

                parent_scope.children.append(new_node)
                if new_node.creates_scope:
                    scope_stack.append(new_node)
                    new_node.enter_scope()

        return root

    def create_node(self, fragment):
        node_class = None
        if fragment.type == TEXT_FRAGMENT:
            node_class = _Text
        elif fragment.type == VAR_FRAGMENT:
            node_class = _Variable
        elif fragment.type == OPEN_BLOCK_FRAGMENT:
            cmd = fragment.clean.split()[0]

            if cmd == 'each':
                node_class = _Each
            elif cmd == 'if':
                node_class = _If
            elif cmd == 'else':
                node_class = _Else
            elif cmd == 'call':
                node_class = _Call
            else:
                node_class = _Block

        if node_class is None:
            raise TemplateSyntaxError(fragment)

        return node_class(fragment.clean)



from pytonik.Log import Log
from pytonik.App import App
from pytonik.Functions.validation import validation
import os, random, string

log_msg = Log()
Ap = App()


class rand(validation):
    def __getattr__(self, item):
        return item

    def __init__(self, *args, **kwargs):
        if len(args) > 0 or len(kwargs) > 0:
            if all(args) != False:
                self.pt = self.number(*args, **kwargs)
            else:
                self.pt = self.number(**kwargs)

        return None

    def __str__(self):

        return self.pt

    def number(self, size=6, list=1):

        num = {
            1: 10,
            2: 120,
            3: 1230,
            4: 12340,
            5: 123450,
            6: 1234560,
            7: 12345670,
            8: 123456780,
            9: 1234567890,
            10: 12345678900,
        }
        ran = []
        for x in range(list):
            ran.append(str(random.randint(size, num[size])))
            
        return "".join(ran) if self.count(ran) < 2 else ran

    def chars(self, size=6, list=1, transform = "lower"):
        transform_s = string.ascii_uppercase if transform == "upper" else string.ascii_lowercase
        strg =  transform_s + string.digits
        ran = []
        for x in range(list):
            ran.append("".join(random.choices(strg, k=size)))

        return "".join(ran) if self.count(ran) < 2 else ran

    def serial(self, size=3, portion=3, sign='-', transform = "lower"):
        transform_s = string.ascii_uppercase if transform == "upper" else string.ascii_lowercase

        strg = transform_s + string.digits
        i = 0
        ran = []
        for x in range(portion):
            ran.append("".join(random.choices(strg, k=size)))

        return str(sign).join(ran)






# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 05/11/2019.
import os

from pytonik.Functions.url import url
from pytonik.util.Variable import Variable
from pytonik import Version

if os.path.isdir(os.getcwd() + '/public'):
    host_public = str(os.getcwd()).replace('\\', '/') + '/public' # os.path.dirname(os.getcwd())
    host_app = str(os.path.dirname(os.getcwd())).replace("\\", "/") + '/app'
else:
    host_public = str(os.getcwd()).replace("\\", "/") 
    host_app = str(os.path.dirname(os.getcwd())).replace("\\", "/")




class path(url, Variable):
    def __getattr__(self, item):
        return item

    def __call__(self, *args, **kwargs):
        
        return None

    def __init__(self, *args, **kwargs):
        
        if len(args) > 0 or len(kwargs) > 0:
            if all(args) is not False:
                self.pt = self.path(*args, **kwargs)
            else:
                self.pt = self.path(**kwargs)

        return None

    def __str__(self):

        return self.pt

    def path(self, path="", link=False):
        
        u = ""
        DS = "/"
        dev_path = ""
        if path[:1] == DS or path[0] == DS:
            DS = ""
        else:
            DS = "/"
        
        if bool(link) == True:
            u = self.url()
        
        return str(u)+str(DS)+str(path)

    def exist(self, newpath, defaultpath="", link=False):
        newpath = "" if newpath == None or newpath == "" else newpath
        path_res = ""
        if os.path.isfile(newpath) == True:
  
            path_res = self.path(path=newpath, link=link)

        elif os.path.isfile(host_app + "/" + str(newpath)) == True:
   
            path_res = self.path(path=newpath, link=link)

        elif os.path.isfile(host_public + "/" + str(newpath)) == True:

            path_res = self.path(path="/public/"+str(+newpath), link=link)

        elif os.path.isdir(newpath) == True:
            path_res =  newpath

        elif os.path.isdir(self.public(path=newpath)) == True:
            path_res =  self.public(path=newpath)

        else:
            
            if defaultpath != "":
                
                if os.path.isfile(defaultpath) == True:
                    path_res = self.path(path=defaultpath, link=link)

                elif os.path.isfile(host_app + "/" + str(defaultpath)) == True:
                    
                    path_res = self.path(path=defaultpath, link=link)

                elif os.path.isfile(host_public + "/" + str(defaultpath)) == True:
                    
                    path_res = self.path(path="/public/"+str(+defaultpath), link=link)

                elif os.path.isdir(defaultpath) == True:
                    path_res =  defaultpath

                elif os.path.isdir(self.public(defaultpath)) == True:
                    
                    path_res =  self.public(path=defaultpath)
        
        return path_res

    def public(self, path):
        DS = "/"
        if path[:1] == DS or path[:1] == DS:
            DS = ""
        else:
            DS = "/"

        return str(host_public) + DS + str(path)



###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###

from pytonik.Editor.Compiler import Compiler

class Template(object):
    def __init__(self, contents):
        self.contents = contents
        self.root = Compiler(self.contents).compile()

    def render(self, **kwargs):
        
        return str(self.root.render(kwargs)).replace('\ufeff', ' ').replace('\u20a6', ' ')


###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###

from pytonik.App import App
from pytonik import Log
from pytonik.Driver.Schema import Schema
import sys
import os
import glob
import importlib
log_msg = Log.Log()


class Model(Schema):

    def __init__(self):
        ap = App()
        self.db = ap.DB() #this is database configuration from module App.py, but it is ignored by Schema.py

    # This Mothod 'load'  load module from model folder
    def load(self, m):

        if os.path.isdir(os.getcwd() + '/public'):
            host = os.getcwd()  # os.path.dirname(os.getcwd())

        else:
            host = os.path.dirname(os.getcwd())

        DS = str("/")

        paths = host + DS + 'model'
        model = paths + DS + m + ".py"
        sys.path.append(paths)
        importlib._RELOADING
        if os.path.isfile(model) == True:

            try:

                md = importlib.import_module(m)
                ob = getattr(md, m)
                if hasattr(ob(), '__call__'):
                    return ob()
                else:
                    log_msg.error("'%s' is not a callable" % m)

            except Exception as err:
                log_msg.error(err)
                self.App.header(0)

        else:

            log_msg.error("Model {e} does not exist ".format(m))



# This function 'auto' Auto load module from model folder
def auto():

    if os.path.isdir(os.getcwd() + '/public'):
        host = os.getcwd()  # os.path.dirname(os.getcwd())

    else:
        host = os.path.dirname(os.getcwd())

    DS = str("/")
    paths = host + DS + 'model'
    status = False
    for file in glob.glob(paths + "/*.py"):

        if os.path.isfile(file) == True:
            name = os.path.splitext(os.path.basename(file))[0]
            # Ignore __ files
            if name.startswith("__init__"):
                continue
            if name != "__init__":
                status = True

    if status == True:
        sys.path.append(paths)



# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 17/12/2019.
#############################################
#############################################
# POSTGRES Support Database Connection
# Using Raw query from psycopg2 module
# Each method represent extensions from POSTGRES version 2.2.9

from pytonik import Log

log_msg = Log.Log()

try:
    import psycopg2
    import psycopg2.extras
except Exception as err:
    log_msg.critical(err)


class pyPgSQL:
   
    def __init__(self, setting):
        self.settings = setting
        self.database = setting.get('database', '') # 
        self.username = setting.get('username', '') # 
        self.password = setting.get('password', '') # 
        self.port = setting.get('port', '') # 
        self.host = setting.get('host', '') # 
        self.prefix = setting.get('prefix', '') # 
        self.Exception = ""
        self.conn =  None
        self.con = None
        self.result = None
        self.connectDB()

    def connectDB(self):

        try:
       
            self.conn = psycopg2.connect(database=self.database, host = self.host,  port = self.port, user=self.username, password=self.password)
       
        except (Exception, psycopg2.Error) as err:
            self.Exception = err
            log_msg.error("Something went wrong : {err}".format(err=err))
            


    def query(self, sql="", value = ""):
        try:
            
            if sql !="" and value != "":
                self.con = self.conn.cursor()
                self.con.execute(str(sql), value)
            else:
                self.con = self.conn.cursor(cursor_factory = psycopg2.extras.RealDictCursor)
                self.con.execute(str(sql))
                
        except (Exception, psycopg2.Error) as err:
            self.Exception = err
            log_msg.error(err)
        return self
        
        
            

    def querymultiple(self, sql="", value = ""):
        try:
            self.con = self.conn.cursor()
            if sql !="" and value != "":
                self.con.executemany(str(sql), value)
            else:
                self.con.executemany(str(sql))
        except (Exception, psycopg2.Error) as err:
            self.Exception = err
            log_msg.error(err)
            
    
        return self


 
    def lastId(self):

        try:
            self.save()
            getrowid = self.con.fetchone()[0]
        except Exception as err:
            getrowid = err

        return getrowid

    def fetch(self):
        result = self.con.fetchall()
        row = []
        for r in result:
            row.append(dict(r))
        return row
       
    def queryone(self, sql="", value = ""):
        self.con = self.conn.cursor()
        if sql !="" and value != "":
            self.con.execute(str(sql), value)
        else:
            self.con.execute(str(sql))

        return self

    def all(self):
        self.result = self.fetch()
        return self.result
    
    def one(self):
        result = self.con.fetchone()
        return dict(result)

    def count(self):
        return self.con.rowcount

    def countall(self):
        self.all()
        return self.con.rowcount

    def save(self):
        try:
            self.conn.commit()
        
            return True
        except Exception as err:
            self.Exception = err
            log_msg.error(err)
            return self

    def close(self):
        return self.con.close()

    def create(self, TABLES = ''):
        self.con = self.conn.cursor()
        if TABLES:
            for table_name in TABLES:
                table_description = TABLES[table_name]
                try:
                    self.con.execute(table_description)
                    self.Exception = "Database table '{}' created successfully.".format(table_name)
                except (Exception, psycopg2.DatabaseError) as err :
                    self.Exception = "Database table '{}' already exists.".format(table_name)
                    log_msg.error("Database table '{}' already exists.".format(table_name))
                
            return self
        
        else:
            self.Exception = "Empty Table"
            return self




# Author : BetaCodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by BetaCodings on 13/01/2020.
#__path__ = __import__('pkgutil').extend_path(__path__, __name__)

import os, sys





###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###

import hashlib
import os
import re
import binascii
from pytonik.Version import *


class Hash:

    def __getattr__(self, item):
        return item

    def __init__(self):
        self.salt = ""
        self.hash_str = ""
        self.salt_hx = ""
        self.type_hx = ""
        self.key_hx = ""
        self.mp_str = ""
        self.hx_mp = {}

    def md5(self, provide_string,  encode='utf-8'):
        if encode == "utf-8":
            hashlibS = hashlib.md5(
                str(provide_string).encode(encode)).hexdigest()
        else:
            hashlibS = hashlib.md5(str(provide_string).encode(
                'utf-8')).hexdigest().encode(encode)

        return hashlibS

    def sha1(self, provide_string,  encode='utf-8'):

        if encode == "utf-8":
            hashlibS = hashlib.sha1(
                str(provide_string).encode(encode)).hexdigest()
        else:
            hashlibS = hashlib.sha1(str(provide_string).encode(
                'utf-8')).hexdigest().encode(encode)

        return hashlibS

    def sha224(self, provide_string,  encode='utf-8'):
        if encode == "utf-8":
            hashlibS = hashlib.sha224(
                str(provide_string).encode(encode)).hexdigest()
        else:
            hashlibS = hashlib.sha224(str(provide_string).encode(
                'utf-8')).hexdigest().encode(encode)
        return hashlibS

    def sha256(self, provide_string,  encode='utf-8'):
        if encode == "utf-8":
            hashlibS = hashlib.sha256(
                str(provide_string).encode(encode)).hexdigest()
        else:
            hashlibS = hashlib.sha256(str(provide_string).encode(
                'utf-8')).hexdigest().encode(encode)
        return hashlibS

    def sha384(self, provide_string, encode='utf-8'):
        if encode == "utf-8":
            hashlibS = hashlib.sha384(
                str(provide_string).encode(encode)).hexdigest()
        else:
            hashlibS = hashlib.sha384(str(provide_string).encode(
                'utf-8')).hexdigest().encode(encode)
        return hashlibS

    def sha512(self, provide_string,  encode='utf-8'):
        if encode == "utf-8":
            hashlibS = hashlib.sha512(
                str(provide_string).encode(encode)).hexdigest()
        else:
            hashlibS = hashlib.sha512(str(provide_string).encode(
                'utf-8')).hexdigest().encode(encode)

        return hashlibS

    def hex_hash(self, provide_string, hex_type='sha256', salt="", size=60):
        
        if salt == "":
            self.salt = self.type_hash(hex_type, os.urandom(int(size)))
            
        else:
            self.salt = salt
        try:
            self.hash_str = hashlib.pbkdf2_hmac(hex_type, str(provide_string).encode('utf-8'),  bytes(self.salt), 100000)
            self.hash_str = binascii.hexlify(self.hash_str)
            return (self.salt+self.hash_str).decode("ascii")
        except Exception as err:
            self.hash_str = hashlib.pbkdf2_hmac(hex_type, str(provide_string).encode('utf-8'),  self.salt.encode('utf-8'), 100000)
            self.hash_str = binascii.hexlify(self.hash_str).decode("ascii")
            return self.hash_str
        
    def type_hash(self, hex_type="sha256", salt="", encode="ascii"):

        salt_re = ""

        if hex_type == "sha256":

            salt_re = self.sha256(provide_string=salt, encode=encode)

        elif hex_type == "sha384":

            salt_re = self.sha384(provide_string=salt, encode=encode)

        elif hex_type == "sha512":

            salt_re = self.sha512(provide_string=salt, encode=encode)

        elif hex_type == "sha1":

            salt_re = self.sha1(provide_string=salt, encode=encode)

        elif hex_type == "md5":

            salt_re = self.md5(provide_string=salt, encode=encode)
        else:
            salt_re = ""

        return salt_re

    def _s_k_(self, stored_hash=""):

        try:
            if PYVERSION_MA >= 3:
                rehx = HASH_PRE.items()
            else :
                rehx = HASH_PRE.iteritems()

            for k, v in rehx:
                len_hash = len(stored_hash)
                
                if int(k) == len_hash:
                    self.type_hx = v
                    self.salt_hx = stored_hash[:int(int(k)/2)]
                    self.key_hx = stored_hash[int(int(k)/2):]
            return True
            
        except Exception as err:
            return False

    def verify(self, stored_hash, provide_string):
        try:
            if self._s_k_(stored_hash) == True:
               
                get_hash = self.hex_hash(provide_string=provide_string, hex_type=self.type_hx, salt=self.salt_hx)
                
                if get_hash == self.key_hx:
                    return True
                else:
                    return False
            else:
                return False
        except Exception as err:
            
            return False

    def map_hash(self, map_string):
        self.mp_str = map_string
        self.hx_mp[map_string] = {
            'salt': self.salt,
            'key': self.hash_str,
        }
        
        return self.hx_mp

    
    def map_check(self, provide_string):
        self.salt_hx = self.hx_mp[self.mp_str]["salt"]
        self.key_hx = self.hx_mp[self.mp_str]["key"]
        self.atr_hx = HASH_PRE.get(str(len(self.salt_hx)+len(self.key_hx)))
        stored_hash = self.hex_hash(provide_string=provide_string, hex_type=self.atr_hx, salt=self.salt_hx)
        return self.verify(stored_hash, provide_string)
   
  




###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###


import importlib
import os
import sys
import re
from pytonik.Editor import Template
from pytonik import Log


log_msg = Log.Log()


class Helpers():

    def __init__(self):
        return None

    def __call__(self, *args, **kwargs):
        return None


if "includeFile" not in dir(os):
    def includeFile(path, datag={}, datal={}):

        host = os.path.dirname(os.getcwd())
        DS = str("/")
        path = '.'+path

        split = path.split('.')

        fileExists = []
        for x in split:

            #fileExists = str(x)
            fileExists.append(x)
        try:
            dirF = list(filter(None, fileExists))
            template_dir = host + DS + 'views' + DS + dirF[0] + DS
            engine = dirF[1]

            if os.path.isdir(template_dir) == True:
                dataG = datag
                dataL = datal
                html_file_path = os.path.join(template_dir, "%s.html" % engine)

                with open(html_file_path) as html_file:
                    html = html_file.read()
                return Template.Template(html).render(**dataG)

            else:
                return "The file {filepath} does not exists.".format(filepath=template_dir)

        except Exception as err:
            log_msg.error(err)
            return err

if "callfunc" not in dir(os):
    def callfunc(myfile, myfunc, *args):
        pathname, filename = os.path.split(myfile)
        sys.path.append(os.path.abspath(pathname))
        modname = os.path.splitext(filename)[0]
        mymod = importlib.import_module(modname)
        try:
            result = getattr(mymod, myfunc)(*args)
            return result
        except Exception as err:
            log_msg.error(err)
            return err


if "url" not in dir(os):

    def url(path="", lang=False):
        from pytonik.Router import Router
        from pytonik.util.Variable import Variable

        ront = Router()
        env_os = Variable()
        DS, p, url = "", "",  ""
        seturl = ront.out("HTTP_HOST")+str(":")+str(env_os.out("SERVER_PORT", '')) if env_os.out(
            "HTTP_HOST") == "localhost" or env_os.out("HTTP_HOST") == "127.0.0.1" else env_os.out("HTTP_HOST")

        if ront.out("HTTPS", "") == 'on':

            url = str("https://") + seturl.replace(":80", "") + "/" + ront.alllanguages.get(ront.getLanguages(), ront.getLanguages()) if lang == True else str(
                "https://") + seturl.replace(":80", "")
        else:
            url = str("http://") + seturl.replace(":80", "") + "/" + ront.alllanguages.get(ront.getLanguages(),
                                                                                           ront.getLanguages()) if lang == True else str("http://") + seturl.replace(":80", "")

        if path == "":
            DS = ""
        else:
            if path[:1] == "/":
                p = path[1:]
                DS = "/"
            else:
                p = path
                DS = "/"

        return url + DS + p

if "public_path" not in dir(os):
    def public_path(path=""):
        DS = str('/')
        return url(DS+'public'+DS+path)


if "covert_list_dict" not in dir(os):
    def covert_list_dict(list):
        itlist = iter(list)
        try:
            return dict(zip(itlist, itlist))
        except Exception as err:
            log_msg.error(err)
            return err


if "mvc_dir" not in dir(os):
    def mvc_dir(path, permission=""):
        dpp = path[len(str(path))-1]
        if '/' in dpp:
            newpath = path[:-1]
        else:
            newpath = path

        if os.path.isdir(os.getcwd() + '/public'):
            host = str(os.getcwd()).replace("\\", "/")  # os.path.dirname(os.getcwd())

        else:
            host = str(os.path.dirname(os.getcwd())).replace("\\", "/") 

        DS = str("/")
        dir_s = host + newpath + DS
        dir_res = ""
        if os.path.exists(dir_s) == True:
            dir_res = dir_s
        else:
            try:
                os.mkdir(dir_s)
                dir_res = dir_s
            except Exception as err:
                try:
                    os.makedirs(dir_s)
                    dir_res = dir_s
                except Exception as err:
                    dir_res = dir_s
        if permission != "":
            try:
                os.chmod(path=dir_s, mode=permission)
            except Exception as err:
                dir_res = dir_s

        return dir_res


if "key_tag" not in dir(os):
    def key_tag(tag, url="/", css=""):

        pattern = re.compile("\s*,\s*|\s+$")

        split_tag = [x for x in pattern.split(tag) if x]
        for keyword in split_tag:
            return '<li><a href="{url}" class="{css}">{keyword}</a></li>'.format(css=css, keyword=str(keyword))

if "alphanumeric" not in dir(os):
    def alphanumeric(string):
        return re.sub("[^a-zA-Z0-9]+", " ", string)


if "rand" not in dir(os):
    def rand(limit=0, list=1):
        import random
        num = {
            1: 10,
            2: 120,
            3: 1230,
            4: 12340,
            5: 123450,
            6: 1234560,
            7: 12345670,
            8: 123456780,
            9: 1234567890,
            10: 12345678900,
        }
        for x in range(list):
            ran = random.randint(limit, num[limit])
            return ran
if "iteration" not in dir(os):

    def iteration(dictionary="", itr="pid"):
        i = 0
        if dictionary != "" or dictionary != None:
            dist, apend = [], []
            for l in dictionary:
                i += 1
                ++i
                listv = l
                dist = {itr: i}
                dist.update(listv)
                apend.append(dist)
            return apend



import re, operator, ast, os, importlib, sys
from pytonik.util.Exception import TemplateSyntaxError, TemplateError, TemplateSyntaxError, TemplateContextError
from pytonik import App, Version

VAR_FRAGMENT = 0
OPEN_BLOCK_FRAGMENT = 1
CLOSE_BLOCK_FRAGMENT = 2
TEXT_FRAGMENT = 3
CLOSE_COMMENT_FRAGMENT = 4

VAR_TOKEN_START = '{{'
VAR_TOKEN_END = '}}'
BLOCK_TOKEN_START = '{%'
BLOCK_TOKEN_END = '%}'
COMMENT_TOKEN_START = '{#'
COMMENT_TOKEN_END = '#}'


#(%s.*?%s|%s.*?%s|%s.*?%s)
TOK_REGEX = (re.compile('(%s.*?%s|%s.*?%s|%s.*?%s)' % (
    re.escape(VAR_TOKEN_START),
    re.escape(VAR_TOKEN_END),
    re.escape(BLOCK_TOKEN_START),
    re.escape(BLOCK_TOKEN_END),
    re.escape(COMMENT_TOKEN_START),
    re.escape(COMMENT_TOKEN_END)
)))

TRANSLATOR_COMMENT_MARK = 'Comment'

WHITESPACE = re.compile("[\s]")
UPPARA = re.compile('\,+')

operator_lookup_table = {
    '<': operator.lt,
    '>': operator.gt,
    '==': operator.eq,
    '!=': operator.ne,
    '<=': operator.le,
    '>=': operator.ge,
    '+': operator.add,
    '-': operator.sub,
    '*': operator.mul,
    '//': operator.floordiv,
    '/': operator.truediv,
    '%': operator.mod,
    '**': operator.pow,
    '<<': operator.lshift,
    '>>': operator.rshift,
    '^': operator.xor,
}

if os.path.isdir(os.getcwd() + '/public'):
    host = os.getcwd()  # os.path.dirname(os.getcwd())

else:
    host = os.path.dirname(os.getcwd())


def eval_expression(expr):

    try:
        return 'literal', ast.literal_eval(expr)
    except Exception as err:
        return 'name', expr


def resolve(name, context):
    
    if name.startswith('..'):
        context = context.get('..', {})
        name = name[2:]
        

    try:
        for tok in name.split('.'):
            context = context[tok]
        return context
    except Exception as err:
        try:
            Ap = App.App()
            load = Ap.loadmodule()
            load.update({name: name})
            md = importlib.import_module(name, name)
            ob = getattr(md, name)
            return ob()
        except Exception as err:
            # name
            raise TemplateContextError(err)

    

def parse_params(params):
        new_args, args, kwargs = [], [], {}

        for param in params:

            if len(params) > 3:

                p = str(params).translate({ord(i): None for i in '"'})

                if '=' in p:

                    new_name = str(p).translate({ord(i): None for i in '"\',\[]'})
                    sln = new_name.split(" ")

                    for n in sln:

                        if '=' in n:

                            k, v = str(n).split("=")

                            kwargs[k] = eval_expression(str(v))
                        else:

                            new_n = str(n).translate({ord(i): None for i in '[]'})
                            new_args.append(new_n)


                    args.append(eval_expression(" ".join(new_args)))

                else:

                    args.append(eval_expression(p))
                break

            else:
                if '=' in param:
                    name, value = param.split('=')

                    kwargs[name] = eval_expression(str(value))
                else:
                    args.append(eval_expression(param))


        return args, kwargs


def dict_local(it, resolves={}):
    l = "" 
    for i, k in enumerate(resolves):
        if k in resolves:
            l = str(it).replace(k, str(resolves[k]))
        if len(resolves) > 1:
            return dict_local_next(l, resolves)
        else:
            return l


def dict_local_next(it, resolves):
    for i, k in enumerate(resolves):
        if k in it:
            it = it.replace(k, resolves[k])
    return it



###
# Author : Betacodings
# Author : info@betacodings.com
# Maintainer By: Emmanuel Martins
# Maintainer Email: emmamartinscm@gmail.com
# Created by Betacodings on 2019.
###


import os, base64, struct, io
from pytonik import App, Log

log_msg = Log.Log()

try:
    from PIL import Image as IMG
except Exception as err:
    log_msg.critical(err)



type_to_content_type = {
    "gif": "image/gif",
    "jpg": "image/jpeg",
    "jpeg": "image/jpeg",
    "pdf": "application/pdf",
    "png": "image/png",
}
type_to_standardized = {
    "gif": "gif",
    "jpg": "jpg",
    "jpeg": "jpg",
    "pdf": "pdf",
    "png": "png",
}


def upload(fileitem, uploaddir, rename=""):
    if os.path.isdir(os.path.dirname(uploaddir)):
        if fileitem.filename != "":
            if rename == "":
                fname = os.path.basename(fileitem.filename)
            else:
                fname = str(rename) + os.path.basename(fileitem.filename)

            try:
                
                with open(str(uploaddir)+str(fname), 'wb') as result:
                    result.write(fileitem.file.read())
                    result.close()
                return True
            except Exception as err:
                log_msg.critical(err)
                return err

        else:
            try:
                
                with open(str(uploaddir)+str(rename), 'wb') as result:
                    result.write(fileitem)
                    result.close()
                return True
            except Exception as err:
                log_msg.critical(err)
                return err
    else:
        log_msg.error("Directory {uploaddir} Does not Exist".format(uploaddir=uploaddir))
        return "Directory {uploaddir} Does not Exist".format(uploaddir=uploaddir)


def delete(directory, file):
    if os.path.isfile(str(directory)+str(file)):
        return os.remove(str(directory)+str(file))
    else:
        log_msg.error("The file {file} does not exist ".format(file=file))
        return "The file {file} does not exist ".format(file=file)

def ext(filename):
    fn = os.path.splitext(filename)[1][1:]
    return fn


class Image():


    
    def __init__(self, dir="", items=""):
        self.s = ""
        self.items = items
        self.dir = dir
        self.blobbase = ""
        self.buffer = ""
        self.pickle = ""
        self.App = App.App()
        self.ext = ""
        self.w = ""
        self.h = ""

        if self.items != "":

            self.read = self.items.file.read()
            self.filename = self.items.filename
        else:
            self.filename = ""
            self.read = ""

    def size(self, image=""):
        
        image = image if image !="" else base64.b64decode(self.blobbase)
        try:
            img = IMG.open(image)
            return img.size
        except Exception as err:
            return err
    
    def byte(self, image=""):
        try:
            return len(self.items.value)
        except Exception as err:
            return err


    def resize(self, width, height, rename=str("")):

        if rename == "":
            fname = os.path.basename(self.items.filename)
        else:
            fname = str(rename) + os.path.basename(self.items.filename)
        try:
            self.blob()
            self.w = width
            self.h = height
            self.ext = ext(self.filename)
            self.filename = str(width) + 'x' + str(height) + '_' + str(fname)
            result = self.creator()
            return True
        except Exception as err:
            log_msg.critical(err)
            return err
        
    def crop(self, dimenion,  image="",  dir_path="", filename=""):
        
        image = image if image !="" else base64.b64decode(self.blobbase)
        filename = filename if filename !="" else self.filename
        dir_path = dir_path if dir_path !="" else self.dir

        try:
            dimenion = dimenion  #(150, 200, 600, 600)
            img = IMG.open(image)
            crop = img.crop(dimenion)
            crop.save(dir_path + filename)
            return True
        except Exception as err:
            log_msg.critical(err)
            return err
    

    def blob(self):
        try:
            self.blobbase = base64.b64encode(self.read)
            return self.blobbase
        except Exception as err:
            log_msg.critical(err)
            return err
            
    def base(self, base_64):
        try:
            base = str(base_64).split(";")
            basev = str(base[1]).split(",")
            self.blobbase = base64.b64decode(basev[1])
            return self.blobbase
        except Exception as err:
            log_msg.critical(err)
            return err
        
    def dimension(self):
        height = -1
        width = -1
        head = self.read
        size = len(head)
        fhandle = self.items.file

        if size >= 10 and head[:6] in (b'GIF87a', b'GIF89a'):

            # Check to see if content_type is correct
            try:
                width, height = struct.unpack("<hh", head[6:10])
            except struct.error:
                raise ValueError("Invalid GIF file")
        # see png edition spec bytes are below chunk length then and finally the
        elif size >= 24 and head.startswith(b'\211PNG\r\n\032\n') and head[12:16] == b'IHDR':

            try:
                width, height = struct.unpack(">LL", head[16:24])
            except struct.error:
                raise ValueError("Invalid PNG file")
        # Maybe this is for an older PNG version.
        elif size >= 16 and head.startswith(b'\211PNG\r\n\032\n'):

            # Check to see if we have the right content type
            try:
                width, height = struct.unpack(">LL", head[8:16])
            except struct.error:
                raise ValueError("Invalid PNG file")
        # handle JPEGs
        elif size >= 2 and head.startswith(b'\377\330'):
            # \xff\xd8\xff\xe0\x00
            try:

                fhandle.seek(0)  # Read 0xff next
                size = 2
                ftype = 0
                while not 0xc0 <= ftype <= 0xcf or ftype in [0xc4, 0xc8, 0xcc]:
                    # fhandle.seek(size, 1)
                    byte = fhandle.read(1)
                    while ord(byte) == 0xff:
                        byte = fhandle.read(1)
                    ftype = ord(byte)
                    size = struct.unpack('>H', fhandle.read(2))[0] - 2
                # We are at a SOFn block
                fhandle.seek(1, 1)  # Skip `precision' byte.
                height, width = struct.unpack('>HH', fhandle.read(4))

            except struct.error:
                raise ValueError("Invalid JPEG file")
        # handle JPEG2000s
        elif size >= 12 and head.startswith(b'\x00\x00\x00\x0cjP  \r\n\x87\n'):
            fhandle.seek(48)

            try:
                height, width = struct.unpack('>LL', fhandle.read(8))
            except struct.error:
                raise ValueError("Invalid JPEG2000 file")


        # handle big endian TIFF
        elif size >= 8 and head.startswith(b"\x4d\x4d\x00\x2a"):
            offset = struct.unpack('>L', head[4:8])[0]
            fhandle.seek(offset)
            ifdsize = struct.unpack(">H", fhandle.read(2))[0]
            for i in range(ifdsize):
                tag, datatype, count, data = struct.unpack(">HHLL", fhandle.read(12))
                if tag == 256:
                    if datatype == 3:
                        width = int(data / 65536)
                    elif datatype == 4:
                        width = data
                    else:
                        raise ValueError("Invalid TIFF file: width column data type should be SHORT/LONG.")
                elif tag == 257:
                    if datatype == 3:
                        height = int(data / 65536)
                    elif datatype == 4:
                        height = data
                    else:
                        raise ValueError("Invalid TIFF file: height column data type should be SHORT/LONG.")
                if width != -1 and height != -1:
                    break
            if width == -1 or height == -1:
                raise ValueError("Invalid TIFF file: width and/or height IDS entries are missing.")
        elif size >= 8 and head.startswith(b"\x49\x49\x2a\x00"):
            offset = struct.unpack('<L', head[4:8])[0]
            fhandle.seek(offset)
            ifdsize = struct.unpack("<H", fhandle.read(2))[0]
            for i in range(ifdsize):
                tag, datatype, count, data = struct.unpack("<HHLL", fhandle.read(12))
                if tag == 256:
                    width = data
                elif tag == 257:
                    height = data
                if width != -1 and height != -1:
                    break
            if width == -1 or height == -1:
                raise ValueError("Invalid TIFF file: width and/or height IDS entries are missing.")

        return width, height

    def show(self, target="image/png"):
        decode = self.blobbase.decode('utf-8')
        data = "data:{};base64,{}".format(target, decode)
        return data

    def open(self):
        if self.file_exist() == True:
            try:

                with open(str(self.dir) + str(self.filename), 'rb') as rb:
                    self.read = rb.read()
                    return self.read
            except Exception as err:
                log_msg.critical(err)
                return err

    def save(self, img_tmp, filename=""):
        filename = filename if filename !="" or filename !=None else self.filename
        try:
            with open(self.dir + filename, 'wb') as result:
                result.write(img_tmp)
                result.close()
            return True
        except Exception as err:
            log_msg.critical(err)
            return err

    def dir_exist(self):
        if os.path.isdir(os.path.dirname(self.dir)):
            return True
        else:
            return False

    def file_exist(self):
        if os.path.isfile(str(self.dir) + str(os.path.basename(self.filename))):
            return True
        else:
            return False

    def upload(self, rename=str(""), uploaddir=""):

        if self.items.filename != "":
            if rename == "":
                fname = os.path.basename(self.items.filename)
            else:
                fname = str(rename)+ str(os.path.basename(self.items.filename))
            
            if uploaddir != "":
                dir_folder = uploaddir
            else:
                dir_folder = self.dir  
            try:

                # result = open(self.dir + fname, 'wb')
                with open(str(dir_folder)+str(fname), 'wb') as result:
                    result.write(self.read)
                    result.close()
                return True
            except Exception as err:
                log_msg.critical(err)
                return err

        else:
            return ""


    
    def creator(self):

        for k in type_to_content_type:
            if self.ext.lower() in k:
                decode = base64.b64decode(self.blobbase)
                try:
                    img = IMG.open(io.BytesIO(decode))
                    resizedIMG = img.resize((self.w, self.h), IMG.ANTIALIAS)
                    resizedIMG.save(self.dir + self.filename, img.format)
                except Exception as err:
                    log_msg.critical(err)
                    return err



from pytonik import App
from pytonik.util.Exception import TemplateSyntaxError, TemplateError, TemplateSyntaxError
from pytonik.Editor.Resolver import *
from pytonik.Log import Log
from pytonik.Version import *

class _Node(object):
    creates_scope = False

    def __init__(self, fragment=None):
        self.children = []
        self.process_fragment(fragment)

    def process_fragment(self, fragment):

        pass

    def enter_scope(self):
        pass

    def render(self, context):

        pass

    def exit_scope(self):
        pass

    def render_children(self, context, children=None):


        if children is None:
            children = self.children

        def render_child(child):

            child_html = child.render(context)

            return '' if not child_html else str(child_html)

        return ''.join(map(render_child, children))


class _ScopableNode(_Node):
    creates_scope = True


class _Root(_Node):
    def render(self, context):

        return self.render_children(context)


class _Variable(_Node):
    def process_fragment(self, fragment):
        self.name = fragment

    def render(self, context):
        
        return resolve(self.name, context)


    

class _Each(_ScopableNode):

    def process_fragment(self, fragment):

        try:

            _, it = WHITESPACE.split(fragment, 1)
            self.it = eval_expression(it)

        except Exception as err:

            raise TemplateSyntaxError(fragment)

    def render(self, context):

        items = self.it[1] if self.it[0] == 'literal' else resolve(self.it[1], context)

        def render_item(item):
            Ap = App.App()
            load = Ap.loadmodule()

            load_m = {'..': context, 'it': item}
            load.update(load_m)

            return self.render_children(load)

        return ''.join(map(render_item, items))

class _Block(_Node):

    def process_fragment(self, fragment):

        try:
            it = WHITESPACE.split(fragment)

            self.block = eval_expression(it)

        except Exception as err:
            raise TemplateSyntaxError(fragment)

    def render(self, context):

        it = self.block[1]

        d, s = [], []
        try:

            for iv in it:
                if iv not in operator_lookup_table:
                    items = resolve(iv, context)
                    d.append(items)
                else:
                    s.append(iv)
            operator_lookup = s[0]
            try:
                return eval(operator_lookup.join(d))

            except Exception as err:
                return operator_lookup.join(d)

        except Exception as err:

            for iv in it:

                if iv not in operator_lookup_table:
                    d.append(iv)
                else:
                    s.append(iv)

            operator_lookup = s[0]
            try:
                return eval(operator_lookup.join(d))
            except Exception as err:
                return operator_lookup.join(d)


class _If(_ScopableNode):
    def process_fragment(self, fragment):
        
        bits = fragment.split()[1:]

        if len(bits) not in (1, 3):
            raise TemplateSyntaxError(fragment)
        self.lhs = eval_expression(bits[0])
        if len(bits) == 3:
            self.op = bits[1]
            
            self.rhs = eval_expression(bits[2])
            
            
            

    def render(self, context):
      
        lhs = self.resolve_side(self.lhs, context)
        
        if hasattr(self, 'op'):
            
            op = operator_lookup_table.get(self.op)
            
            if op is None:
                raise TemplateSyntaxError(self.op)

            
            
            rhs = self.resolve_side(self.rhs, context)
            
            exec_if_branch = op(lhs, rhs)
        else:
            exec_if_branch = operator.truth(lhs)


        self.if_branch, self.else_branch = self.split_children()
        return self.render_children(context, self.if_branch if exec_if_branch else self.else_branch)

    def resolve_side(self, side, context):
        
        return side[1] if side[0] == 'literal' else resolve(side[1], context)
        

    def exit_scope(self):
        self.if_branch, self.else_branch = self.split_children()

    def split_children(self):
        if_branch, else_branch = [], []
        curr = if_branch
        for child in self.children:
            if isinstance(child, _Else):
                curr = else_branch
                continue
            curr.append(child)
        return if_branch, else_branch


class _Else(_Node):
    def render(self, context):
        pass



class _Call(_Node):

    def process_fragment(self, fragment):
        
        try:
            #bits = WHITESPACE.
            self.bits = WHITESPACE.split(fragment)
            self.callable = self.bits[1]
            self.fragment = fragment
            self.args, self.kwargs = parse_params(self.bits[2:])

        except Exception as err:
            raise TemplateSyntaxError(fragment)

    
    def render(self, context):
        
        self.contxt = context

        ob_dir = [str(os.path.dirname(__file__).replace('Editor', '')) + str("Functions"), str(host) + str("/") + "model"]

        resolved_args, resolved_kwargs = [], {}

        for kind, value in self.args:
            if kind == 'name':
                value = value

            
            value = self._call_each(str(value))
            
            resolved_args.append(value)
            
        if PYVERSION_MA >= 2:
            items  = self.kwargs.items()
        else:
            items = self.kwargs.iteritems()

        for key, (kind, value) in items:
            
            if kind == 'name':
                value = value
            value = self._call_each(str(value))
            try:
                valux = eval(value)
            except Exception as err:
                
                valux = value
            resolved_kwargs[key] = valux

        path = [str(ob_dir[0]) + "/" + str(self.callable) + ".py",  str(ob_dir[1]) + "/" + str(self.callable) + ".py"]

        sys.path.append(str(ob_dir[0]))
        sys.path.append(str(ob_dir[1]))


        importlib._RELOADING
        
        if os.path.isfile(path[0]) == True:
            md = importlib.import_module(self.callable, self.callable)
            ob = getattr(md, self.callable)

            if hasattr(ob(), '__call__') == True:

                calls = ""
                try:
                    newob = getattr(ob, ''.join(resolved_args).replace(" ", ""))
                    calls = newob(**resolved_kwargs)
                    
                except Exception as err:
                    try:
                        newob = getattr(ob(), ''.join(resolved_args).replace(" ", ""))
                        calls = newob(**resolved_kwargs)
                    except Exception as err:
                        try:
                            _cal = ob()
                            _new_cal = getattr(_cal, *resolved_args)
                            calls =  _new_cal(**resolved_kwargs)

                        except Exception as err:
                            try:
                                calls = ob(*resolved_args, **resolved_kwargs)
                            except Exception as err:
                                Log('').error(err)

                return calls
            else:
                raise TemplateError("'%s' is not a callable" % self.callable)

        elif os.path.isfile(path[1]) == True:


            md = importlib.import_module(self.callable, self.callable)
            ob = getattr(md, self.callable)


            if hasattr(ob(), '__call__'):

                calls = ""
                try:
                    newob = getattr(ob, ''.join(resolved_args).replace(" ", ""))
                    calls = newob(**resolved_kwargs)
                except Exception as err:

                    try:
                        newob = getattr(ob(), ''.join(resolved_args).replace(" ", ""))
                        calls = newob(**resolved_kwargs)
                    except Exception as err:
                        try:
                            _cal = ob()
                            _new_cal = getattr(_cal, *resolved_args)
                            calls = _new_cal(**resolved_kwargs)

                        except Exception as err:

                            try:
                                calls = ob(*resolved_args, **resolved_kwargs)

                            except Exception as err:
                                Log('').error(err)
                return calls
            else:
                raise TemplateError("'%s' is not a callable" % self.callable)

        else:
            raise TemplateError("'%s' module not found " % self.callable)

    def _call_each(self, context):

        if  VAR_TOKEN_START in context:
            self.it = str(context).translate({ord(i): None for i in '{VAR_TOKEN_START}{VAR_TOKEN_END}'.format(VAR_TOKEN_START = '{{', VAR_TOKEN_END = '}}')})
            oparatork = "/"
            for oparator_k,  oparator_v in operator_lookup_table.items():
                if oparator_k in self.it:
                    oparatork = oparator_k
            contsplit = self.it.split(oparatork)
            
            if len(contsplit) < 2:
                item = ""
                for k, itc in enumerate(contsplit):

                    if 'it.' in itc:

                        ite = resolve(itc, self.contxt)
                        item = str(self.it).replace(itc, str(ite))
            
                    elif '..' in itc:
                        ite = resolve(itc, self.contxt)
                        item = str(self.it).replace(itc, str(ite))
                        
                    else:
                        try:
                            ite = resolve(itc, self.contxt)
                            item = str(self.it).replace(itc, str(ite))
                        except Exception as err:
                            err = ""
                return item

            elif len(contsplit) > 1:
                
                dic_ls = {}
                dict_r = ""
                
                for s in contsplit:

                    if 'it.'.lower() in s.lower():
                        dic_ls.update({s: resolve(s, self.contxt)})
                        dict_r = dict_local(self.it, dic_ls)

                    elif '..'.lower() in s.lower():
                        dic_ls.update({s: resolve(s, self.contxt)})
                        dict_r = dict_local(self.it, dic_ls)
                    else:
                        
                        try:
                            dic_ls.update({s: resolve(s, self.contxt)})
                            dict_r = dict_local(self.it, dic_ls)
                        except Exception as err:
                            err = ""
                        
                return  dict_r

            else:
                item = ""
                if 'it.' in self.it:
                    ite = resolve(self.it, self.contxt)
                    item = str(self.it).replace(self.it, ite)
                elif '..' in self.it:

                    ite = resolve(self.it, self.contxt)
                    item = str(self.it).replace(self.it, ite)
                else:
                    try:
                        ite = resolve(self.it, self.contxt)
                        item = str(self.it).replace(self.it, ite)
                    except Exception as err:
                        err = ""
                        
                return item

        elif '..' in context:
            item = ""
            for it in context.split('/'):
                if '..' in it:
                    ite = resolve(it, self.contxt)
                    item = str(context).replace(it, ite)
                else:
                    try:
                        ite = resolve(it, self.contxt)
                        item = str(context).replace(it, ite)
                    except Exception as err:
                        err = ""
            return item
        else:
            return context



class _Text(_Node):
    def process_fragment(self, fragment):
        self.text = fragment
    def render(self, context):
        
        return self.text
